/*
 * FactorOracleAutomaton.h
 *
 *  Created on: Jan 8, 2017
 *      Author: Jan Travnicek
 */

#ifndef FACTOR_ORACLE_AUTOMATON_H_
#define FACTOR_ORACLE_AUTOMATON_H_

#include <alib/string>
#include <alib/set>
#include <alib/iostream>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>
#include <primitive/Unsigned.h>
#include <primitive/xml/Unsigned.h>

#include <alphabet/common/SymbolNormalize.h>

#include <automaton/FSM/DFA.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class FactorOracleAutomaton final : public object::ObjectBase {
protected:
	automaton::DFA < SymbolType, unsigned > m_automaton;

public:
	/**
	 * @copydoc OracleTrieNode::clone ( ) const &
	 */
	virtual ObjectBase * clone ( ) const &;

	/**
	 * @copydoc OracleTrieNode::clone ( ) const &
	 */
	virtual ObjectBase * clone ( ) &&;

	explicit FactorOracleAutomaton ( automaton::DFA < SymbolType, unsigned > automaton );

	/**
	 * @return Root node of the trie
	 */
	const automaton::DFA < SymbolType, unsigned > & getAutomaton ( ) const &;

	automaton::DFA < SymbolType, unsigned > && getAutomaton ( ) &&;

	const ext::set < SymbolType > & getAlphabet ( ) const & {
		return m_automaton.getInputAlphabet ( );
	}

	ext::set < SymbolType > && getAlphabet ( ) && {
		return std::move ( m_automaton ).getInputAlphabet ( );
	}

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const SymbolType & symbol ) {
		return m_automaton.removeInputSymbol ( symbol );
	}

	unsigned getBackboneLength ( ) const {
		return m_automaton.getStates ( ).size ( ) - 1;
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const FactorOracleAutomaton & other ) const;

	virtual explicit operator std::string ( ) const;

	explicit operator automaton::DFA < SymbolType, unsigned > ( ) const;

	virtual object::ObjectBase * inc ( ) &&;

	typedef FactorOracleAutomaton < > normalized_type;
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType >
FactorOracleAutomaton < SymbolType >::FactorOracleAutomaton ( automaton::DFA < SymbolType, unsigned > automaton ) : m_automaton ( std::move ( automaton ) ) {
}

template < class SymbolType >
object::ObjectBase * FactorOracleAutomaton < SymbolType >::clone ( ) const & {
	return new FactorOracleAutomaton ( * this );
}

template < class SymbolType >
object::ObjectBase * FactorOracleAutomaton < SymbolType >::clone ( ) && {
	return new FactorOracleAutomaton ( std::move ( * this ) );
}

template < class SymbolType >
const automaton::DFA < SymbolType, unsigned > & FactorOracleAutomaton < SymbolType >::getAutomaton ( ) const & {
	return m_automaton;
}

template < class SymbolType >
automaton::DFA < SymbolType, unsigned > && FactorOracleAutomaton < SymbolType >::getAutomaton ( ) && {
	return std::move ( m_automaton );
}

template < class SymbolType >
void FactorOracleAutomaton < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(FactorOracleAutomaton " << this->m_automaton << ")";
}

template < class SymbolType >
int FactorOracleAutomaton < SymbolType >::compare ( const FactorOracleAutomaton & other ) const {
	auto first = ext::tie ( getAutomaton ( ) );
	auto second = ext::tie ( other.getAutomaton ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
FactorOracleAutomaton < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
FactorOracleAutomaton < SymbolType >::operator automaton::DFA < SymbolType, unsigned > ( ) const {
	return getAutomaton ( );
}

template < class SymbolType >
object::ObjectBase* FactorOracleAutomaton < SymbolType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace core {

template < class SymbolType >
struct normalize < indexes::stringology::FactorOracleAutomaton < SymbolType >, typename std::enable_if < ! std::is_same < indexes::stringology::FactorOracleAutomaton < SymbolType >, indexes::stringology::FactorOracleAutomaton < > >::value >::type > {
	static indexes::stringology::FactorOracleAutomaton < > eval ( indexes::stringology::FactorOracleAutomaton < SymbolType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( std::move ( value ).getAutomaton ( ) ).getInputAlphabet ( ) );
		ext::set < unsigned > states = std::move ( std::move ( value ).getAutomaton ( ) ).getStates ( );
		unsigned initialState = std::move ( std::move ( value ).getAutomaton ( ) ).getInitialState ( );
		ext::set < unsigned > finalStates = std::move ( std::move ( value ).getAutomaton ( ) ).getFinalStates ( );

		automaton::DFA < DefaultStateType, unsigned > res ( std::move ( states ), std::move ( alphabet ), std::move ( initialState ), std::move ( finalStates ) );

		for ( std::pair < ext::pair < unsigned, SymbolType >, unsigned > && transition : ext::make_moveable_map ( std::move ( std::move ( value ).getAutomaton ( ) ).getTransitions ( ) ) ) {
			unsigned from = transition.first.first;
			DefaultSymbolType input = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( transition.first.second ) );
			unsigned to = transition.second;

			res.addTransition ( std::move ( from ), std::move ( input ), std::move ( to ) );
		}

		return indexes::stringology::FactorOracleAutomaton < > ( std::move ( res ) );
	}
};

template < class SymbolType >
struct xmlApi < indexes::stringology::FactorOracleAutomaton < SymbolType > > {
	static indexes::stringology::FactorOracleAutomaton < SymbolType > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const indexes::stringology::FactorOracleAutomaton < SymbolType > & data );
};

template < class SymbolType >
indexes::stringology::FactorOracleAutomaton < SymbolType > xmlApi < indexes::stringology::FactorOracleAutomaton < SymbolType > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	automaton::DFA < SymbolType, unsigned > automaton = core::xmlApi < automaton::DFA < SymbolType, unsigned > >::parse ( input );
	indexes::stringology::FactorOracleAutomaton < SymbolType > res ( std::move ( automaton ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return res;
}

template < class SymbolType >
bool xmlApi < indexes::stringology::FactorOracleAutomaton < SymbolType > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < class SymbolType >
const std::string & xmlApi < indexes::stringology::FactorOracleAutomaton < SymbolType > >::xmlTagName ( ) {
	static std::string xmlTagName = "FactorOracleAutomaton";

	return xmlTagName;
}

template < class SymbolType >
void xmlApi < indexes::stringology::FactorOracleAutomaton < SymbolType > >::compose ( ext::deque < sax::Token > & output, const indexes::stringology::FactorOracleAutomaton < SymbolType > & index ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	core::xmlApi < automaton::DFA < SymbolType, unsigned > >::compose ( output, index.getAutomaton ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace core */

#endif /* FACTOR_ORACLE_AUTOMATON_H_ */
